import logging
from pathlib import Path
from typing import Any, Callable, Optional, TypedDict

from wpextract.download.exceptions import WordPressApiNotV2
from wpextract.download.exporter import Exporter
from wpextract.download.requestsession import HTTPError, RequestSession
from wpextract.download.wpapi import WPApi, WPObject

ExportCallable = Callable[[list[WPObject], Path], int]


class _ObjTypeFetchData(TypedDict):
    export_func: ExportCallable
    obj_name: str


class WPDownloader:
    """Manages the download of data from a WordPress site."""

    def __init__(
        self,
        target: str,
        out_path: Path,
        data_types: list[str],
        session: Optional[RequestSession] = None,
        json_prefix: Optional[str] = None,
    ) -> None:
        """Initializes the WPDownloader object.

        Args:
            target: the target WordPress site URL
            out_path: the output path for the downloaded data
            data_types: set of data types to download
            session: request session. Will be created from default constructor if not provided.
            json_prefix: prefix to prepend to JSON file names
        """
        self.target = target
        self.out_path = out_path
        self.data_types = data_types
        self.session = session if session else RequestSession()
        self._test_session()
        self.scanner = WPApi(self.target, session=self.session)
        self.json_prefix = json_prefix
        self.media_cache: Optional[list[WPObject]] = None

    def _test_session(self) -> None:
        try:
            self.session.get(self.target)
            logging.info("Connected successfully")
        except Exception as e:
            logging.error("Failed to connect to the server")
            raise e

    def download(self) -> None:
        """Download and export the requested data lists."""
        if "users" in self.data_types:
            self._list_obj(WPApi.USER)
        if "tags" in self.data_types:
            self._list_obj(WPApi.TAG)
        if "categories" in self.data_types:
            self._list_obj(WPApi.CATEGORY)
        if "posts" in self.data_types:
            self._list_obj(WPApi.POST)
        if "pages" in self.data_types:
            self._list_obj(WPApi.PAGE)
        if "comments" in self.data_types:
            self._list_obj(WPApi.COMMENT)
        if "media" in self.data_types:
            self._list_obj(WPApi.MEDIA)

    def download_media_files(self, session: RequestSession, dest: Path) -> None:
        """Download site media files.

        Args:
            session: the request session to use
            dest: destination directory for media
        """
        logging.info("Pulling media URLs")
        media, slugs = self.scanner.get_media_urls("all", media_cache=self.media_cache)

        if len(media) == 0:
            logging.warning("No media found corresponding to the criteria")
            return
        logging.info(f"{len(media)} media URLs found")

        number_dl = Exporter.download_media(session, media, dest)
        logging.info(f"Downloaded {number_dl} media files")

    def _get_fetch_or_list_type(
        self, obj_type: int, plural: bool = False
    ) -> _ObjTypeFetchData:
        """Returns a dict containing all necessary metadata about the obj_type to list and fetch data.

        Args:
            obj_type: the type of the object
            plural: whether the name must be plural or not

        Raises:
            ValueError: if the object type is unknown

        Returns:
            A dict containing the export function (`export_func`) and the object name (`obj_name`)
        """
        export_func: Optional[ExportCallable] = None
        obj_name = ""
        if obj_type == WPApi.USER:
            export_func = Exporter.export_users
            obj_name = "Users" if plural else "User"
        elif obj_type == WPApi.TAG:
            export_func = Exporter.export_tags
            obj_name = "Tags" if plural else "Tag"
        elif obj_type == WPApi.CATEGORY:
            export_func = Exporter.export_categories
            obj_name = "Categories" if plural else "Category"
        elif obj_type == WPApi.POST:
            export_func = Exporter.export_posts
            obj_name = "Posts" if plural else "Post"
        elif obj_type == WPApi.PAGE:
            export_func = Exporter.export_pages
            obj_name = "Pages" if plural else "Page"
        elif obj_type == WPApi.COMMENT:
            export_func = Exporter.export_comments_interactive
            obj_name = "Comments" if plural else "Comment"
        elif obj_type == WPApi.MEDIA:
            export_func = Exporter.export_media
            obj_name = "Media"
        else:
            raise ValueError(f"Unknown object type {obj_type}")

        return {
            "export_func": export_func,
            "obj_name": obj_name,
        }

    def _list_obj(
        self,
        obj_type: int,
        start: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> None:
        prop = self._get_fetch_or_list_type(obj_type, plural=True)
        logging.info(f"Downloading {prop['obj_name']}")

        try:
            obj_list, _ = self.scanner.get_obj_list(obj_type, start, limit)

            WPDownloader.export_decorator(
                export_func=prop["export_func"],
                file_name=prop["obj_name"].lower(),
                json_path=self.out_path,
                json_prefix=self.json_prefix,
                values=obj_list,
            )
            if obj_type == WPApi.MEDIA:
                self.media_cache = obj_list
        except HTTPError:
            logging.exception(
                f"An HTTP error was encountered while downloading {prop['obj_name']}"
            )
        except WordPressApiNotV2:
            logging.error("The API does not support WP V2")
        except OSError as e:
            logging.error(f"Could not open {e.filename} for writing")
        logging.info(f"Completed downloading {prop['obj_name']}")

    @staticmethod
    def export_decorator(
        export_func: ExportCallable,
        file_name: str,
        json_path: Path,
        json_prefix: Optional[str],
        values: Any,
        kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """Call the export function with a constructed filename.

        Args:
            export_func: the function to run
            file_name: the name of the export file
            json_path: the path to the output directory
            json_prefix: a prefix for the export file
            values: data to be exported
            kwargs: arguments to pass to the export function
        """
        kwargs = kwargs or {}

        filename = file_name + ".json"
        if json_prefix is not None:
            filename = json_prefix + "-" + filename

        json_file = json_path / filename
        export_func(values, json_file, **kwargs)



import copy
import html
import json
from pathlib import Path
from typing import Any, Optional, Union
from urllib import parse as urlparse

from tqdm.auto import tqdm

from wpextract.download.requestsession import RequestSession


class Exporter:
    """Utility functions to export data."""

    CHUNK_SIZE = 2048
    """The size of chunks to download large files"""

    @staticmethod
    def download_media(
        session: RequestSession, media: list[str], out_path: Path
    ) -> int:
        """Downloads the media files based on the given URLs.

        Args:
            session: the request session to use
            media: the URLs as a list
            out_path: the path to the folder where the files are being saved, it is assumed as existing

        Returns:
            the number of files written
        """
        files_number = 0
        for m in tqdm(media, unit="media"):
            r = session.do_request("get", m, stream=True)
            if r.status_code == 200:
                http_path = urlparse.urlparse(m).path.split("/")
                local_path = out_path
                if len(http_path) > 1:
                    for el in http_path[:-1]:
                        local_path = local_path / el
                        local_path.mkdir(exist_ok=True)
                local_path = local_path / http_path[-1]
                with open(local_path, "wb") as f:
                    i = 0
                    content_size = int(r.headers.get("Content-Length", -1))
                    chunks_pbar = None
                    if content_size > 0:
                        chunks_pbar = tqdm(
                            total=content_size,
                            unit="B",
                            unit_scale=True,
                            desc=http_path[-1],
                            leave=False,
                        )
                    for chunk in r.iter_content(Exporter.CHUNK_SIZE):
                        if chunks_pbar is not None:
                            chunks_pbar.update(Exporter.CHUNK_SIZE)
                        f.write(chunk)
                        i += 1
                    if chunks_pbar is not None:
                        chunks_pbar.update(content_size - chunks_pbar.n)
                        chunks_pbar.close()
                files_number += 1
        return files_number

    @staticmethod
    def setup_export(
        vlist: list[dict[str, Any]], parameters_to_unescape: list[Union[str, list[str]]]
    ) -> list[dict[str, Any]]:
        """Sets up the right values for a list export.

        This function flattens a list of objects before its serialization in the expected format.
        It also makes a deepcopy to ensure that the original vlist is not altered.

        Args:
            vlist: the list to prepare for exporting
            parameters_to_unescape: parameters to unescape (ex.
                ["param1", ["param2"]["rendered"]])

        Returns:
            the list of objects ready to be exported
        """
        exported_list = []

        for el in vlist:
            if el is not None:
                # First copy the object
                exported_el = copy.deepcopy(el)
                # Look for parameters to HTML unescape
                for key in parameters_to_unescape:
                    if type(key) is str:  # If the parameter is at the root
                        exported_el[key] = html.unescape(exported_el[key])
                    elif type(key) is list:  # If the parameter is nested
                        selected: Optional[Union[dict[str, Any], str]] = exported_el
                        siblings = []
                        fullpath: dict[str, Any] = {}
                        # We look for the leaf first, not forgetting sibling branches for rebuilding the tree later
                        for k in key:
                            if type(selected) is dict and k in selected.keys():
                                sib = {}
                                for e in selected.keys():
                                    if e != k:
                                        sib[e] = selected[e]
                                selected = selected[k]
                                siblings.append(sib)
                            else:
                                selected = None
                                break
                        # If we can unescape the parameter, we do it and rebuild the tree starting from the leaf
                        if selected is not None and type(selected) is str:
                            selected = html.unescape(selected)
                            key.reverse()
                            fullpath[key[0]] = selected
                            s = len(siblings) - 1
                            for e in siblings[s].keys():
                                fullpath[e] = siblings[s][e]
                            for k in key[1:]:
                                fullpath = {k: fullpath}
                                s -= 1
                                for e in siblings[s].keys():
                                    fullpath[e] = siblings[s][e]
                            key.reverse()
                            exported_el[key[0]] = fullpath[key[0]]
                # The resulting element is appended to the list of exported elements
                exported_list.append(exported_el)

        return exported_list

    @staticmethod
    def write_file(filename: Path, data: Any) -> None:
        """Writes content to the given file in JSON format.

        The key mapping must be a dict of keys or lists of keys to ensure proper mapping.

        Args:
            filename: the path of the file
            data: the actual data to export
        """
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

    @staticmethod
    def export_posts(
        posts: list[dict[str, Any]],
        filename: Path,
    ) -> int:
        """Exports posts to the specified file.

        Args:
            posts: the posts to export
            filename: filename to use

        Returns:
            the length of the list written to the file
        """
        exported_posts = Exporter.setup_export(
            posts,
            [["title", "rendered"], ["content", "rendered"], ["excerpt", "rendered"]],
        )

        Exporter.write_file(filename, exported_posts)
        return len(exported_posts)

    @staticmethod
    def export_categories(categories: list[dict[str, Any]], filename: Path) -> int:
        """Exports categories to the specified file.

        Args:
            categories: the categories to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_categories = Exporter.setup_export(
            categories,  # TODO
            [],
        )

        Exporter.write_file(filename, exported_categories)
        return len(exported_categories)

    @staticmethod
    def export_tags(tags: list[dict[str, Any]], filename: Path) -> int:
        """Exports tags to the specified file.

        Args:
            tags: the tags to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_tags = tags  # It seems that no modification will be done for this one, so no deepcopy
        Exporter.write_file(filename, exported_tags)
        return len(exported_tags)

    @staticmethod
    def export_users(users: list[dict[str, Any]], filename: Path) -> int:
        """Exports users to the specified file.

        Args:
            users: the users to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_users = users  # It seems that no modification will be done for this one, so no deepcopy
        Exporter.write_file(filename, exported_users)
        return len(exported_users)

    @staticmethod
    def export_pages(pages: list[dict[str, Any]], filename: Path) -> int:
        """Exports pages to the specified file.

        Args:
            pages: the pages to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_pages = Exporter.setup_export(
            pages,
            [
                ["guid", "rendered"],
                ["title", "rendered"],
                ["content", "rendered"],
                ["excerpt", "rendered"],
            ],
        )

        Exporter.write_file(filename, exported_pages)
        return len(exported_pages)

    @staticmethod
    def export_media(media: list[dict[str, Any]], filename: Path) -> int:
        """Exports media to the specified file.

        Args:
            media: the media to export
            filename: file to export to

        Returns:
            the length of the list written to the file
        """
        exported_media = Exporter.setup_export(
            media,
            [
                ["guid", "rendered"],
                ["title", "rendered"],
                ["description", "rendered"],
                ["caption", "rendered"],
            ],
        )

        Exporter.write_file(filename, exported_media)
        return len(exported_media)

    # FIXME to be refactored
    @staticmethod
    def export_comments_interactive(
        comments: list[dict[str, Any]], filename: Path
    ) -> int:
        """Exports comments to the specified file.

        Args:
            comments: the comments to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_comments = Exporter.setup_export(
            comments,
            [["content", "rendered"]],
        )

        Exporter.write_file(filename, exported_comments)
        return len(exported_comments)



import logging
import random
import time
from http.cookies import SimpleCookie
from typing import TYPE_CHECKING, Literal, Optional, Union

import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from urllib3 import Retry

if TYPE_CHECKING:
    from requests.models import Response
    from requests.sessions import _Data as RequestDataType

DEFAULT_UA = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36"


class ConnectionCouldNotResolve(Exception):
    """The remote host could not be resolved."""

    pass


class ConnectionReset(Exception):
    """The connection was reset during the request."""

    pass


class ConnectionRefused(Exception):
    """The connection was refused by the server."""

    pass


class ConnectionTimeout(Exception):
    """A connection timeout occurred."""

    pass


class HTTPError(Exception):
    """Base class for HTTP errors."""

    pass


class HTTPError400(HTTPError):
    """HTTP Bad Request.

    See Also:
        HTTPErrorInvalidPage for a special case of this error.
    """

    pass


class HTTPErrorInvalidPage(HTTPError):
    """Special case of HTTP 400 if the error is caused by a nonexistent page.

    This indicates the last page has been passed and all items have been retrieved.
    """

    pass


class HTTPError401(HTTPError):
    """HTTP Unauthorized."""

    pass


class HTTPError403(HTTPError):
    """HTTP Forbidden."""

    pass


class HTTPError404(HTTPError):
    """HTTP Not Found."""

    pass


class HTTPError500(HTTPError):
    """HTTP Internal Server Error."""

    pass


class HTTPError502(HTTPError):
    """HTTP Bad Gateway."""

    pass


class HTTPTooManyRedirects(Exception):
    """Raised if the number of allowed redirects exceeds the configured maximum value."""

    pass


RETRY_AFTER_STATUS = frozenset(
    [
        # Standard HTTP
        413,  # Payload Too Large
        429,  # Too Many Requests
        500,  # Internal Server Error
        502,  # Bad Gateway
        503,  # Service Unavailable
        504,  # Gateway Timeout,
        # Cloudflare
        520,  # Web Server Returned an Unknown Error
        522,  # Connection Timed Out
        524,  # A Timeout Occurred
    ]
)

ERROR_NAMES = {
    400: "Bad Request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not Found",
    500: "Internal Server Error",
    502: "Bad Gateway",
}
EXCEPTION_CLS = {
    400: HTTPError400,
    401: HTTPError401,
    403: HTTPError403,
    404: HTTPError404,
    500: HTTPError500,
    502: HTTPError502,
}


def _handle_status(url: str, status_code: int, n_tries: Optional[int] = None) -> None:
    if status_code in ERROR_NAMES:
        log_msg = (
            f'Error {status_code} ({ERROR_NAMES[status_code]}) while fetching "{url}"'
        )
    else:
        log_msg = f'Error {status_code} while fetching "{url}"'

    if n_tries is not None:
        log_msg += f" after {n_tries} tries"

    if status_code in EXCEPTION_CLS or status_code > 400:
        if status_code == 404:
            logging.debug(log_msg)
        else:
            logging.error(log_msg)

        if status_code in EXCEPTION_CLS:
            raise EXCEPTION_CLS[status_code]
        else:
            raise HTTPError


class RequestWait:
    """Manages waiting between requests."""

    def __init__(
        self, wait: Optional[float] = None, random_wait: Optional[bool] = False
    ):
        """Create a new waiting instance.

        Args:
            wait: The base time to wait in seconds
            random_wait: If true, the wait duration will be multiplied by a random factor between 0.5 and 1.5.
        """
        self.wait_s = wait or 0
        self.random_wait = random_wait

    def wait(self) -> None:
        """Perform the specified wait."""
        if self.wait is None:
            return

        wait_factor = 1.0
        if self.random_wait:
            wait_factor = random.uniform(0.5, 1.5)

        time.sleep(self.wait_s * wait_factor)


AuthorizationType = Union[tuple[str, str], HTTPBasicAuth, HTTPDigestAuth]


class RequestSession:
    """Manages HTTP requests and their behaviour."""

    def __init__(
        self,
        proxy: Optional[str] = None,
        cookies: Optional[str] = None,
        authorization: Optional[AuthorizationType] = None,
        timeout: Optional[float] = 30,
        wait: Optional[float] = None,
        random_wait: bool = False,
        max_retries: int = 10,
        backoff_factor: float = 0.1,
        max_redirects: int = 20,
        user_agent: Optional[str] = None,
    ):
        """Create a new request session.

        Args:
            proxy: a dict containing a proxy server string for HTTP and/or HTTPS connection
            cookies: a string in the format of the Cookie header
            authorization: a tuple containing login and password or [`requests.auth.HTTPBasicAuth`][requests.auth.HTTPBasicAuth] for basic authentication or [`requests.auth.HTTPDigestAuth`][requests.auth.HTTPDigestAuth] for NTLM-like authentication
            timeout: maximum time in seconds to wait for a response before giving up
            wait: wait time in seconds between requests, None to not wait
            random_wait: If true, the wait time between requests is multiplied by a random factor between 0.5 and 1.5
            max_retries: the maximum number of retries before failing
            backoff_factor: Factor to wait between successive retries
            max_redirects: maximum number of redirects to follow
            user_agent: User agent to use for requests. Set to [`DEFAULT_UA`][wpextract.download.requestsession.DEFAULT_UA] by default.
        """
        self.s = requests.Session()
        if proxy is not None:
            self.set_proxy(proxy)
        if cookies is not None:
            self.set_cookies(cookies)
        if authorization is not None and (
            (type(authorization) is tuple and len(authorization) == 2)
            or type(authorization) is HTTPBasicAuth
            or type(authorization) is HTTPDigestAuth
        ):
            self.s.auth = authorization
        self.wait = wait
        self.timeout = timeout
        self.s.max_redirects = max_redirects
        self._mount_retry(backoff_factor, max_retries)
        self.waiter = RequestWait(wait, random_wait)
        self.user_agent = user_agent if user_agent is not None else DEFAULT_UA

    def _mount_retry(self, backoff_factor: float, max_retries: int) -> None:
        retry = Retry(
            total=max_retries,
            backoff_factor=backoff_factor,
            status_forcelist=RETRY_AFTER_STATUS,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.s.mount("http://", adapter)
        self.s.mount("https://", adapter)

    def get(self, url: str) -> "Response":
        """Calls the get function from requests but handles errors to raise proper exception following the context.

        Args:
            url: URL to fetch

        Returns:
            the Response object
        """
        return self.do_request("get", url)

    def post(self, url: str, data: Optional["RequestDataType"] = None) -> "Response":
        """Calls the post function from requests but handles errors to raise proper exception following the context.

        Args:
            url: URL to fetch
            data: optional data to send

        Returns:
            the Response object
        """
        return self.do_request("post", url, data)

    def do_request(
        self,
        method: Literal["get", "post"],
        url: str,
        data: Optional["RequestDataType"] = None,
        stream: bool = False,
    ) -> "Response":
        """Helper class to regroup requests and handle exceptions at the same location.

        Args:
            method: HTTP method to use
            url: URL to fetch
            data: optional data to send
            stream: if True, the response will be streamed

        Raises:
            ConnectionCouldNotResolve: The remote host could not be resolved.
            ConnectionRefused: The connection was refused by the server.
            ConnectionReset: The connection was reset during the request.
            ConnectionTimeout: Exceeded the configured timeout
            HTTPErrorInvalidPage: Special case of HTTP 400 if the error is caused by a nonexistent page.
                This usually signals that available pages have been exhausted.

        Returns:
            the Response object
        """
        headers = {"User-Agent": self.user_agent}
        response = None
        try:
            if method == "post":
                response = self.s.post(url, data, headers=headers, timeout=self.timeout)
            else:
                response = self.s.get(
                    url, headers=headers, timeout=self.timeout, stream=stream
                )
        except requests.ConnectionError as e:
            if "Errno -5" in str(e) or "Errno -2" in str(e) or "Errno -3" in str(e):
                logging.error(f"Could not resolve host {url}")
                raise ConnectionCouldNotResolve from e
            elif "Errno 111" in str(e):
                logging.error(f"Connection refused by {url}")
                raise ConnectionRefused from e
            elif "RemoteDisconnected" in str(e):
                logging.error(f"Connection reset by {url}")
                raise ConnectionReset from e
            else:
                raise e
        except requests.Timeout as e:
            logging.error(f"Request timed out fetching {url}")
            raise ConnectionTimeout from e
        except requests.TooManyRedirects as e:
            logging.error(f'Too many redirects while fetching "{url}"')
            raise HTTPTooManyRedirects from e

        # If this is an HTTP 400 due to an invalid page, raise this special error early
        if response.status_code == 400 and "application/json" in response.headers.get(
            "content-type", ""
        ):
            json_body = response.json()
            if "code" in json_body and "invalid_page_number" in json_body["code"]:
                raise HTTPErrorInvalidPage

        n_tries = None
        if hasattr(response.raw, "retries") and response.raw.retries is not None:
            # initial request + n retires = n+1 tries
            n_tries = len(response.raw.retries.history) + 1

        _handle_status(url, response.status_code, n_tries)

        self.waiter.wait()
        return response

    def set_cookies(self, cookies: str) -> None:
        """Sets new cookies from a string.

        Args:
            cookies: Cookies in a format usable by [http.cookies.SimpleCookie.load][]
        """
        c = SimpleCookie()
        c.load(cookies)
        for key, m in c.items():
            self.s.cookies.set(key, m.value)

    def get_cookies(self) -> dict[str, str]:
        """Retrieve cookies in the cookie jar.

        Returns:
            Cookies in dictionary format.
        """
        return self.s.cookies.get_dict()

    def set_proxy(self, proxy: str) -> None:
        """Set a proxy to use for the request.

        Args:
            proxy: proxy URL In the format supported by requests
        """
        prot = "http"
        if proxy[:5].lower() == "https":
            prot = "https"
        self.s.proxies = {prot: proxy}

    def set_creds(self, credentials: AuthorizationType) -> None:
        """Set new credentials for the request.

        Args:
            credentials: credentials as an HTTPBasic tuple or supported requests HTTP auth instance
        """
        self.s.auth = credentials



import copy
import logging
import math
from json.decoder import JSONDecodeError
from typing import Any, Literal, Optional, Union

from tqdm.auto import tqdm

from wpextract.download.exceptions import (
    NoWordpressApi,
    NSNotFoundException,
    WordPressApiNotV2,
)
from wpextract.download.requestsession import (
    HTTPError,
    HTTPError404,
    HTTPErrorInvalidPage,
    RequestSession,
)
from wpextract.download.utils import (
    get_by_id,
    get_content_as_json,
    url_path_join,
)

WPObject = dict[str, Any]
ObjectsAndTotal = tuple[list[WPObject], Optional[int]]


class WPApi:
    """Queries the WordPress API to retrieve information."""

    # Object types
    POST = 0
    """The post type"""
    POST_REVISION = 1
    """The post revision type"""
    WP_BLOCK = 2
    """The Gutenberg block type"""
    CATEGORY = 3
    """The category type"""
    TAG = 4
    """The tag type"""
    PAGE = 5
    """The page type"""
    COMMENT = 6
    """The comment type"""
    MEDIA = 7
    """The media type"""
    USER = 8
    """The user type"""
    THEME = 9
    """The theme type"""
    NAMESPACE = 10
    """The namespace type"""
    # SEARCH_RESULT = 10
    ALL_TYPES = 20
    """Constant representing all types"""

    def __init__(
        self,
        target: str,
        api_path: str = "wp-json/",
        session: Optional[RequestSession] = None,
    ) -> None:
        """Creates a new instance of WPApi.

        Args:
            target: the target of the scan
            api_path: the api path, if non-default
            session: the requests session object to use for HTTP requests
        """
        self.api_path = api_path
        self.has_v2: Optional[bool] = None
        self.name = None
        self.description = None
        self.url = target
        self.basic_info: Optional[dict[str, Any]] = None

        if session is not None:
            self.s = session
        else:
            self.s = RequestSession()

    def get_basic_info(self) -> dict[Any, Any]:
        """Collects and stores basic information about the target.

        Raises:
            NoWordpressApi: The target does not have a reachable WordPress API

        Returns:
            The basic information about the target
        """
        rest_url = url_path_join(self.url, self.api_path)
        if self.basic_info is not None:
            return self.basic_info

        try:
            req = self.s.get(rest_url)
        except Exception as e:
            raise NoWordpressApi from e
        if req.status_code >= 400:
            raise NoWordpressApi

        try:
            self.basic_info = get_content_as_json(req)
        except JSONDecodeError as e:
            raise NoWordpressApi from e

        if "name" in self.basic_info.keys():
            self.name = self.basic_info["name"]

        if "description" in self.basic_info.keys():
            self.description = self.basic_info["description"]

        if (
            "namespaces" in self.basic_info.keys()
            and "wp/v2" in self.basic_info["namespaces"]
        ):
            self.has_v2 = True

        return self.basic_info

    def crawl_pages(
        self,
        url: str,
        start: Optional[int] = None,
        num: Optional[int] = None,
        display_progress: bool = True,
    ) -> tuple[list[WPObject], int]:
        """Crawls all pages while there is at least one result for the given endpoint or tries to get pages from start to end.

        Args:
            url: the URL to crawl
            start: the start index
            num: the number of entries to retrieve
            display_progress: whether to display a progress bar

        Raises:
            WordPressApiNotV2: The target does not support the WordPress API v2
            HTTPError: An HTTP error is encountered before any content is retrieved

        Returns:
            A tuple containing the list of entries and the total number of entries
        """
        page = 1
        total_entries = 0
        total_pages = 0
        more_entries = True
        entries: list[WPObject] = []
        base_url = url
        entries_left = 1
        per_page = 10
        if start is not None:
            page = math.floor(start / per_page) + 1
        if num is not None:
            entries_left = num

        # Initialise placeholder for progress bar
        pbar = None

        while more_entries and entries_left > 0:
            rest_url = url_path_join(self.url, self.api_path, (base_url % page))
            if start is not None:
                rest_url += "&per_page=%d" % per_page
            try:
                req = self.s.get(rest_url)
                if (
                    page == 1
                    or (start is not None and page == math.floor(start / per_page) + 1)
                ) and "X-WP-Total" in req.headers:
                    total_entries = int(req.headers["X-WP-Total"])
                    total_pages = int(req.headers["X-WP-TotalPages"])
                    logging.info("Total number of entries: %d" % total_entries)
                    if start is not None and total_entries < start:
                        start = total_entries - 1
            except HTTPErrorInvalidPage:
                logging.debug(
                    "Received HTTP 400 error which appears to be an invalid page error, probably reached the end."
                )
                break
            except HTTPError as e:
                if len(entries) == 0:
                    raise e

                logging.exception(
                    f"Error while fetching page {page}. Stopping at {len(entries)} entries."
                )
                break
            except Exception as e:
                logging.error(f"Error while fetching page {page}.")
                raise e

            try:
                json_content = get_content_as_json(req)
                if type(json_content) is list and len(json_content) > 0:
                    if (
                        start is None
                        or (
                            start is not None
                            and page > math.floor(start / per_page) + 1
                        )
                    ) and num is None:
                        entries += json_content
                        if start is not None:
                            entries_left -= len(json_content)
                    elif start is not None and page == math.floor(start / per_page) + 1:
                        first_idx = start % per_page
                        if num is None or (
                            num is not None and len(json_content[first_idx:]) < num
                        ):
                            entries += json_content[first_idx:]
                            if num is not None:
                                entries_left -= len(json_content[first_idx:])
                        else:
                            entries += json_content[first_idx : first_idx + num]
                            entries_left = 0
                    else:
                        if num is not None and entries_left > len(json_content):
                            entries += json_content
                            entries_left -= len(json_content)
                        else:
                            entries += json_content[:entries_left]
                            entries_left = 0

                    if display_progress:
                        if num is None and start is None and total_entries >= 0:
                            if pbar is None:
                                pbar = tqdm(total=total_pages)
                            pbar.update(page - pbar.n)

                        elif num is None and start is not None and total_entries >= 0:
                            if pbar is None:
                                pbar = tqdm(total=total_entries - start)
                            pbar.update(total_entries - start - entries_left - pbar.n)
                        elif num is not None and total_entries > 0:
                            if pbar is None:
                                pbar = tqdm(total=total_entries)
                            pbar.update(num - entries_left - pbar.n)
                else:
                    more_entries = False
            except JSONDecodeError:
                more_entries = False

            page += 1

        if pbar is not None:
            pbar.close()

        return (entries, total_entries)

    def crawl_single_page(self, url: str) -> Any:
        """Crawls a single URL.

        Args:
            url: the URL to crawl

        Raises:
            WordPressApiNotV2: The target does not support the WordPress API v2

        Returns:
            The content of the page
        """
        content = None
        rest_url = url_path_join(self.url, self.api_path, url)
        try:
            req = self.s.get(rest_url)
        except HTTPErrorInvalidPage:
            return None
        except HTTPError404:
            return None

        try:
            content = get_content_as_json(req)
        except JSONDecodeError:
            pass

        return content

    def get_comments(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all comments.

        Args:
            start: the start index
            num: the number of entries to retrieve

        Returns:
            The list of comments and total number of comments available
        """
        return self.crawl_pages("wp/v2/comments?page=%d", start, num)

    def get_posts(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all posts.

        Args:
            comments: whether to retrieve comments
            start: the start index
            num: the number of entries to retrieve

        Raises:
            WordPressApiNotV2: The target does not support the WordPress API v2

        Returns:
            The list of posts and total number of posts available
        """
        if self.has_v2 is None:
            self.get_basic_info()
        if not self.has_v2:
            raise WordPressApiNotV2

        return self.crawl_pages("wp/v2/posts?page=%d", start=start, num=num)

    def get_tags(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all tags.

        Args:
            start: the start index
            num: the number of entries to retrieve

        Returns:
            The list of tags and total number of tags available
        """
        return self.crawl_pages("wp/v2/tags?page=%d", start, num)

    def get_categories(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all categories.

        Args:
            start: the start index
            num: the number of entries to retrieve

        Returns:
            The list of categories and total number of categories available
        """
        return self.crawl_pages("wp/v2/categories?page=%d", start=start, num=num)

    def get_users(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all users.

        Args:
            start: the start index
            num: the number of entries to retrieve

        Returns:
            The list of users and total number of users available
        """
        return self.crawl_pages("wp/v2/users?page=%d", start=start, num=num)

    def get_media(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all media objects.

        Args:
            start: the start index
            num: the number of entries to retrieve

        Returns:
            The list of media objects
        """
        return self.crawl_pages("wp/v2/media?page=%d", start=start, num=num)

    def get_media_urls(
        self,
        ids: Union[Literal["all"], str],
        media_cache: Optional[list[WPObject]] = None,
    ) -> tuple[list[str], list[str]]:
        """Retrieves the media download URLs for specified IDs or all or from cache.

        Args:
            ids: the IDs of the media objects to retrieve, "all" for all, "cache" for cached, or a comma-separated list of IDs
            media_cache: a list of previously retrieved media to use

        Returns:
            A tuple containing the list of URLs and the list of slugs
        """
        media: list[WPObject] = []
        if ids == "all":
            if media_cache is None:
                media, _ = self.get_media()
            else:
                media = media_cache
        else:
            id_list = ids.split(",")
            media = []
            for i in id_list:
                try:
                    if int(i) > 0:
                        m = self.get_obj_by_id(
                            WPApi.MEDIA,
                            int(i),
                            media_cache,
                        )
                        if m is not None and len(m) > 0 and type(m[0]) is dict:
                            media.append(m[0])
                except ValueError:
                    pass
        urls = []
        slugs = []
        if media is None:
            return [], []
        for m_item in media:
            if (
                m_item is not None
                and type(m_item) is dict
                and "source_url" in m_item.keys()
                and "slug" in m_item.keys()
            ):
                urls.append(m_item["source_url"])
                slugs.append(m_item["slug"])
        return urls, slugs

    def get_pages(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> ObjectsAndTotal:
        """Retrieves all pages.

        Args:
            start: the start index
            num: the number of entries to retrieve
            force: ignore cache and force a re-fetch

        Returns:
            The list of pages
        """
        return self.crawl_pages("wp/v2/pages?page=%d", start=start, num=num)

    def get_namespaces(
        self,
        start: Optional[int] = None,
        num: Optional[int] = None,
    ) -> list[str]:
        """Retrieves an array of namespaces.

        Args:
            start: the start index
            num: the number of entries to retrieve
            force: ignore cache and force a re-fetch

        Returns:
            The list of namespaces
        """
        if self.has_v2 is None:
            self.get_basic_info()
        if self.basic_info is None:
            return []
        if "namespaces" in self.basic_info.keys():
            if start is None and num is None:
                return self.basic_info["namespaces"]  # type: ignore[no-any-return]
            namespaces = copy.deepcopy(self.basic_info["namespaces"])
            if start is not None and start < len(namespaces):
                namespaces = namespaces[start:]
            if num is not None and num <= len(namespaces):
                namespaces = namespaces[:num]
            return namespaces  # type: ignore[no-any-return]
        return []

    def get_routes(self) -> dict[str, Any]:
        """Retrieves a dictionary of routes.

        Returns:
            The dictionary of routes
        """
        if self.has_v2 is None:
            self.get_basic_info()
        if self.basic_info is None:
            return {}
        if "routes" in self.basic_info.keys():
            return self.basic_info["routes"]  # type: ignore[no-any-return]
        return {}

    def crawl_namespaces(self, ns: Union[Literal["all"], str]) -> dict[str, Any]:
        """Crawls all accessible get routes defined for the specified namespace.

        Args:
            ns: the namespace to crawl, or "all" for all namespaces

        Raises:
            NSNotFoundException: If a namespace was specified but not found

        Returns:
            A dictionary containing the data for the specified namespace
        """
        namespaces = self.get_namespaces()
        routes = self.get_routes()
        ns_data = {}
        if ns != "all" and ns not in namespaces:
            raise NSNotFoundException
        for url, route in routes.items():
            if "namespace" not in route.keys() or "endpoints" not in route.keys():
                continue
            url_as_ns = url.lstrip("/")
            if "(?P<" in url or url_as_ns in namespaces:
                continue
            if (ns != "all" and route["namespace"] != ns) or route["namespace"] in [
                "wp/v2",
                "",
            ]:
                continue
            for endpoint in route["endpoints"]:
                if "GET" not in endpoint["methods"]:
                    continue
                keep = True
                if len(endpoint["args"]) > 0 and type(endpoint["args"]) is dict:
                    for name, arg in endpoint["args"].items():  # noqa: B007
                        if arg["required"]:
                            keep = False
                if keep:
                    rest_url = url_path_join(self.url, self.api_path, url)
                    try:
                        ns_request = self.s.get(rest_url)
                        ns_data[url] = get_content_as_json(ns_request)
                    except Exception:
                        continue
        return ns_data

    def get_obj_by_id_helper(
        self, obj_id: int, url: str, cache: list[WPObject]
    ) -> list[WPObject]:
        """Retrieve an object from the cache or get it if not present.

        Args:
            cache: list of objects to use as cache
            obj_id: id of the object to fetch
            url: URL formatting template containing "%d" where the ID should be substituted
            use_cache: whether to use the cache or force a re-fetch

        Returns:
            A list containing the returned object, empty if not retrievable.
        """
        obj = get_by_id(cache, obj_id)
        if obj is not None:
            return [obj]

        obj = self.crawl_single_page(url % obj_id)
        if type(obj) is dict:
            return [obj]
        return []

    def get_obj_by_id(
        self, obj_type: int, obj_id: int, obj_cache: Optional[list[WPObject]]
    ) -> list[WPObject]:
        """Returns a list of maximum one object specified by its type and ID.

        Also returns an empty list if the ID does not exist.

        Args:
            obj_type: the type of the object (ex. POST)
            obj_id: the ID of the object to fetch
            obj_cache: a list of cached objects

        Returns:
            A list containing the returned object, empty if not retrievable.
        """
        obj_cache = [] if obj_cache is None else obj_cache

        if obj_type == WPApi.USER:
            return self.get_obj_by_id_helper(obj_id, "wp/v2/users/%d", obj_cache)
        if obj_type == WPApi.TAG:
            return self.get_obj_by_id_helper(obj_id, "wp/v2/tags/%d", obj_cache)
        if obj_type == WPApi.CATEGORY:
            return self.get_obj_by_id_helper(
                obj_id,
                "wp/v2/categories/%d",
                obj_cache,
            )
        if obj_type == WPApi.POST:
            return self.get_obj_by_id_helper(
                obj_id,
                "wp/v2/posts/%d",
                obj_cache,
            )
        if obj_type == WPApi.PAGE:
            return self.get_obj_by_id_helper(
                obj_id,
                "wp/v2/pages/%d",
                obj_cache,
            )
        if obj_type == WPApi.COMMENT:
            return self.get_obj_by_id_helper(
                obj_id,
                "wp/v2/comments/%d",
                obj_cache,
            )
        if obj_type == WPApi.MEDIA:
            return self.get_obj_by_id_helper(
                obj_id,
                "wp/v2/media/%d",
                obj_cache,
            )
        return []

    def get_obj_list(
        self,
        obj_type: int,
        start: Optional[int],
        limit: Optional[int],
    ) -> ObjectsAndTotal:
        """Returns a list of maximum limit objects specified by the starting object offset.

        Args:
            obj_type: the type of the object (ex. POST)
            start: the offset of the first object to return
            limit: the maximum number of objects to return


        Returns:
            A list of the returned objects
        """
        get_func = None
        if obj_type == WPApi.USER:
            get_func = self.get_users
        elif obj_type == WPApi.TAG:
            get_func = self.get_tags
        elif obj_type == WPApi.CATEGORY:
            get_func = self.get_categories
        elif obj_type == WPApi.PAGE:
            get_func = self.get_pages
        elif obj_type == WPApi.POST:
            get_func = self.get_posts
        elif obj_type == WPApi.COMMENT:
            get_func = self.get_comments
        elif obj_type == WPApi.MEDIA:
            get_func = self.get_media
        elif obj_type == WPApi.NAMESPACE:
            get_func = self.get_namespaces  # type: ignore[assignment]

        if get_func is not None:
            return get_func(start=start, num=limit)

        return [], None
